import enum
import json
import os
import time
import requests
import numpy as np

from ophyd import EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV
from ophyd import Device, Staged
from ophyd import ADComponent as ADCpt

from bec_lib import messages, MessageEndpoints, bec_logger

from ophyd_devices.epics.devices.psi_detector_base import PSIDetectorBase, CustomDetectorMixin

logger = bec_logger.logger

MIN_READOUT = 3e-3


class PilatusError(Exception):
    """Base class for exceptions in this module."""


class PilatusTimeoutError(PilatusError):
    """Raised when the Pilatus does not respond in time during unstage."""


class TriggerSource(enum.IntEnum):
    """Trigger source options for the detector"""

    INTERNAL = 0
    EXT_ENABLE = 1
    EXT_TRIGGER = 2
    MULTI_TRIGGER = 3
    ALGINMENT = 4


class SLSDetectorCam(Device):
    """SLS Detector Camera - Pilatus

    Base class to map EPICS PVs to ophyd signals.
    """

    num_images = ADCpt(EpicsSignalWithRBV, "NumImages")
    num_frames = ADCpt(EpicsSignalWithRBV, "NumExposures")
    delay_time = ADCpt(EpicsSignalWithRBV, "NumExposures")
    trigger_mode = ADCpt(EpicsSignalWithRBV, "TriggerMode")
    acquire = ADCpt(EpicsSignal, "Acquire")
    armed = ADCpt(EpicsSignalRO, "Armed")

    read_file_timeout = ADCpt(EpicsSignal, "ImageFileTmot")
    detector_state = ADCpt(EpicsSignalRO, "StatusMessage_RBV")
    status_message_camserver = ADCpt(EpicsSignalRO, "StringFromServer_RBV", string=True)
    acquire_time = ADCpt(EpicsSignal, "AcquireTime")
    acquire_period = ADCpt(EpicsSignal, "AcquirePeriod")
    threshold_energy = ADCpt(EpicsSignalWithRBV, "ThresholdEnergy")
    file_path = ADCpt(EpicsSignalWithRBV, "FilePath")
    file_name = ADCpt(EpicsSignalWithRBV, "FileName")
    file_number = ADCpt(EpicsSignalWithRBV, "FileNumber")
    auto_increment = ADCpt(EpicsSignalWithRBV, "AutoIncrement")
    file_template = ADCpt(EpicsSignalWithRBV, "FileTemplate")
    file_format = ADCpt(EpicsSignalWithRBV, "FileNumber")
    gap_fill = ADCpt(EpicsSignalWithRBV, "GapFill")


class PilatusSetup(CustomDetectorMixin):
    """Pilatus setup class for cSAXS

    Parent class: CustomDetectorMixin

    """

    def initialize_default_parameter(self) -> None:
        """Set default parameters for Eiger9M detector"""
        self.update_readout_time()

    def update_readout_time(self) -> None:
        """Set readout time for Eiger9M detector"""
        readout_time = (
            self.parent.scaninfo.readout_time
            if hasattr(self.parent.scaninfo, "readout_time")
            else self.parent.MIN_READOUT
        )
        self.parent.readout_time = max(readout_time, self.parent.MIN_READOUT)

    def initialize_detector(self) -> None:
        """Initialize detector"""
        # Stops the detector
        self.stop_detector()
        # Sets the trigger source to GATING
        self.parent.set_trigger(TriggerSource.EXT_ENABLE)

    def prepare_detector(self) -> None:
        """
        Prepare detector for scan.

        Includes checking the detector threshold,
        setting the acquisition parameters and setting the trigger source
        """
        self.set_detector_threshold()
        self.set_acquisition_params()
        self.parent.set_trigger(TriggerSource.EXT_ENABLE)

    def set_detector_threshold(self) -> None:
        """
        Set correct detector threshold to 1/2 of current X-ray energy, allow 5% tolerance

        Threshold might be in ev or keV
        """

        # get current beam energy from device manageer
        mokev = self.parent.device_manager.devices.mokev.obj.read()[
            self.parent.device_manager.devices.mokev.name
        ]["value"]
        factor = 1

        # Check if energies are eV or keV, assume keV as the default
        unit = getattr(self.parent.cam.threshold_energy, "units", None)
        if unit is not None and unit == "eV":
            factor = 1000

        # set energy on detector
        setpoint = int(mokev * factor)

        # set threshold on detector
        threshold = self.parent.cam.threshold_energy.read()[self.parent.cam.threshold_energy.name][
            "value"
        ]
        if not np.isclose(setpoint / 2, threshold, rtol=0.05):
            self.parent.cam.threshold_energy.set(setpoint / 2)

    def set_acquisition_params(self) -> None:
        """Set acquisition parameters for the detector"""

        # Set number of images and frames (frames is for internal burst of detector)
        self.parent.cam.num_images.put(
            int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger)
        )
        self.parent.cam.num_frames.put(1)

        # Update the readout time of the detector
        self.update_readout_time()

    def create_directory(self, filepath: str) -> None:
        """Create directory if it does not exist"""
        os.makedirs(filepath, exist_ok=True)

    def stop_detector_backend(self) -> None:
        """Stop the file writer zmq service for pilatus_2"""
        self.close_file_writer()
        time.sleep(0.1)
        self.stop_file_writer()
        time.sleep(0.1)

    def close_file_writer(self) -> None:
        """
        Close the file writer for pilatus_2

        Delete the data from x12sa-pd-2

        """
        url = "http://x12sa-pd-2:8080/stream/pilatus_2"
        try:
            res = self.send_requests_delete(url=url)
            if not res.ok:
                res.raise_for_status()
        except Exception as exc:
            logger.info(f"Pilatus2 close threw Exception: {exc}")

    def stop_file_writer(self) -> None:
        """
        Stop the file writer for pilatus_2

        Runs on xbl-daq-34
        """
        url = "http://xbl-daq-34:8091/pilatus_2/stop"
        res = self.send_requests_put(url=url)
        if not res.ok:
            res.raise_for_status()

    def prepare_data_backend(self) -> None:
        """
        Prepare the detector backend of pilatus for a scan

        A zmq service is running on xbl-daq-34 that is waiting
        for a zmq message to start the writer for the pilatus_2 x12sa-pd-2

        """

        self.stop_detector_backend()

        self.parent.filepath = self.parent.filewriter.compile_full_filename(
            self.parent.scaninfo.scan_number, "pilatus_2.h5", 1000, 5, True
        )
        self.parent.cam.file_path.put("/dev/shm/zmq/")
        self.parent.cam.file_name.put(
            f"{self.parent.scaninfo.username}_2_{self.parent.scaninfo.scan_number:05d}"
        )
        self.parent.cam.auto_increment.put(1)  # auto increment
        self.parent.cam.file_number.put(0)  # first iter
        self.parent.cam.file_format.put(0)  # 0: TIFF
        self.parent.cam.file_template.put("%s%s_%5.5d.cbf")

        # TODO better to remove hard coded path with link to home directory/pilatus_2
        basepath = f"/sls/X12SA/data/{self.parent.scaninfo.username}/Data10/pilatus_2/"
        self.parent.filepath_raw = os.path.join(
            basepath,
            self.parent.filewriter.get_scan_directory(self.parent.scaninfo.scan_number, 1000, 5),
        )
        # Make directory if needed
        self.create_directory(self.parent.filepath_raw)

        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        # start the stream on x12sa-pd-2
        url = "http://x12sa-pd-2:8080/stream/pilatus_2"
        data_msg = {
            "source": [
                {
                    "searchPath": "/",
                    "searchPattern": "glob:*.cbf",
                    "destinationPath": self.parent.filepath_raw,
                }
            ]
        }
        res = self.send_requests_put(url=url, data=data_msg, headers=headers)
        logger.info(f"{res.status_code} -  {res.text} - {res.content}")

        if not res.ok:
            res.raise_for_status()

        # start the data receiver on xbl-daq-34
        url = "http://xbl-daq-34:8091/pilatus_2/run"
        data_msg = [
            "zmqWriter",
            self.parent.scaninfo.username,
            {
                "addr": "tcp://x12sa-pd-2:8888",
                "dst": ["file"],
                "numFrm": int(
                    self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger
                ),
                "timeout": 2000,
                "ifType": "PULL",
                "user": self.parent.scaninfo.username,
            },
        ]
        res = self.send_requests_put(url=url, data=data_msg, headers=headers)
        logger.info(f"{res.status_code}  - {res.text} - {res.content}")

        if not res.ok:
            res.raise_for_status()

        # Wait for server to become available again
        time.sleep(0.1)
        logger.info(f"{res.status_code} -{res.text} - {res.content}")

        # Send requests.put to xbl-daq-34 to wait for data
        url = "http://xbl-daq-34:8091/pilatus_2/wait"
        data_msg = [
            "zmqWriter",
            self.parent.scaninfo.username,
            {
                "frmCnt": int(
                    self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger
                ),
                "timeout": 2000,
            },
        ]
        try:
            res = self.send_requests_put(url=url, data=data_msg, headers=headers)
            logger.info(f"{res}")

            if not res.ok:
                res.raise_for_status()
        except Exception as exc:
            logger.info(f"Pilatus2 wait threw Exception: {exc}")

    def send_requests_put(self, url: str, data: list = None, headers: dict = None) -> object:
        """
        Send a put request to the given url

        Args:
            url (str): url to send the request to
            data (dict): data to be sent with the request (optional)
            headers (dict): headers to be sent with the request (optional)

        Returns:
            status code of the request
        """
        return requests.put(url=url, data=json.dumps(data), headers=headers, timeout=5)

    def send_requests_delete(self, url: str, headers: dict = None) -> object:
        """
        Send a delete request to the given url

        Args:
            url (str): url to send the request to
            headers (dict): headers to be sent with the request (optional)

        Returns:
            status code of the request
        """
        return requests.delete(url=url, headers=headers, timeout=5)

    def pre_scan(self) -> None:
        """
        Pre_scan function call

        This function is called just before the scan core.
        Here it is used to arm the detector for the acquisition

        """
        self.arm_acquisition()

    def arm_acquisition(self) -> None:
        """Arms the detector for the acquisition"""
        self.parent.cam.acquire.put(1)
        # TODO is this sleep needed? to be tested with detector and for how long
        time.sleep(0.5)

    def publish_file_location(self, done: bool = False, successful: bool = None) -> None:
        """
        Publish the filepath to REDIS and publish the event for the h5_converter

        We publish two events here:
        - file_event: event for the filewriter
        - public_file: event for any secondary service (e.g. radial integ code)

        Args:
            done (bool): True if scan is finished
            successful (bool): True if scan was successful
        """
        pipe = self.parent.connector.pipeline()
        if successful is None:
            msg = messages.FileMessage(
                file_path=self.parent.filepath,
                done=done,
                metadata={"input_path": self.parent.filepath_raw},
            )
        else:
            msg = messages.FileMessage(
                file_path=self.parent.filepath,
                done=done,
                successful=successful,
                metadata={"input_path": self.parent.filepath_raw},
            )
        self.parent.connector.set_and_publish(
            MessageEndpoints.public_file(self.parent.scaninfo.scanID, self.parent.name),
            msg,
            pipe=pipe,
        )
        self.parent.connector.set_and_publish(
            MessageEndpoints.file_event(self.parent.name), msg, pipe=pipe
        )
        pipe.execute()

    def finished(self) -> None:
        """Check if acquisition is finished."""
        # pylint: disable=protected-access
        # TODO: at the moment this relies on device.mcs.obj._staged attribute
        signal_conditions = [
            (
                lambda: self.parent.device_manager.devices.mcs.obj._staged,
                Staged.no,
            ),
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=True,
        ):
            raise PilatusTimeoutError(
                f"Reached timeout with detector state {signal_conditions[0][0]}, std_daq state"
                f" {signal_conditions[1][0]} and received frames of {signal_conditions[2][0]} for"
                " the file writer"
            )
        self.stop_detector()
        self.stop_detector_backend()

    def stop_detector(self) -> None:
        """Stop detector"""
        self.parent.cam.acquire.put(0)

    def check_scanID(self) -> None:
        """Checks if scanID has changed and stops the scan if it has"""
        old_scanID = self.parent.scaninfo.scanID
        self.parent.scaninfo.load_scan_metadata()
        if self.parent.scaninfo.scanID != old_scanID:
            self.parent.stopped = True


class PilatuscSAXS(PSIDetectorBase):
    """Pilatus_2 300k detector for CSAXS

    Parent class: PSIDetectorBase

    class attributes:
        custom_prepare_cls (Eiger9MSetup)   : Custom detector setup class for cSAXS,
                                              inherits from CustomDetectorMixin
        cam (SLSDetectorCam)                : Detector camera
        MIN_READOUT (float)                 : Minimum readout time for the detector

    """

    # Specify which functions are revealed to the user in BEC client
    USER_ACCESS = [
        "describe",
    ]

    # specify Setup class
    custom_prepare_cls = PilatusSetup
    # specify minimum readout time for detector
    MIN_READOUT = 3e-3
    # specify class attributes
    cam = ADCpt(SLSDetectorCam, "cam1:")

    def set_trigger(self, trigger_source: TriggerSource) -> None:
        """Set trigger source for the detector"""
        value = trigger_source
        self.cam.trigger_mode.put(value)


if __name__ == "__main__":
    pilatus_2 = PilatuscSAXS(name="pilatus_2", prefix="X12SA-ES-PILATUS300K:", sim_mode=True)



from .sim import (
    SimPositioner,
    SimMonitor,
    SimCamera,
    SynDynamicComponents,
    SynFlyer,
)
from .sim_xtreme import SynXtremeOtf

from .sim_signals import SetableSignal, ReadOnlySignal, ComputedReadOnlySignal



import enum
import threading
from collections import defaultdict
import numpy as np

from ophyd import EpicsSignal, EpicsSignalRO, Device, Component as Cpt
from ophyd_devices.utils import bec_utils
from ophyd_devices.epics.devices.psi_detector_base import PSIDetectorBase, CustomDetectorMixin

from bec_lib import messages, MessageEndpoints, bec_logger, threadlocked


logger = bec_logger.logger


class MCSError(Exception):
    """Base class for exceptions in this module."""


class MCSTimeoutError(MCSError):
    """Raise when MCS card runs into a timeout"""


class TriggerSource(int, enum.Enum):
    """Trigger source for mcs card - see manual for more information"""

    MODE0 = 0
    MODE1 = 1
    MODE2 = 2
    MODE3 = 3
    MODE4 = 4
    MODE5 = 5
    MODE6 = 6


class ChannelAdvance(int, enum.Enum):
    """Channel advance pixel mode for mcs card - see manual for more information"""

    INTERNAL = 0
    EXTERNAL = 1


class ReadoutMode(int, enum.Enum):
    """Readout mode for mcs card - see manual for more information"""

    PASSIVE = 0
    EVENT = 1
    IO_INTR = 2
    FREQ_0_1HZ = 3
    FREQ_0_2HZ = 4
    FREQ_0_5HZ = 5
    FREQ_1HZ = 6
    FREQ_2HZ = 7
    FREQ_5HZ = 8
    FREQ_10HZ = 9
    FREQ_100HZ = 10


class MCSSetup(CustomDetectorMixin):
    """Setup mixin class for the MCS card"""

    def __init__(self, *args, parent: Device = None, **kwargs) -> None:
        super().__init__(*args, parent=parent, **kwargs)
        self._lock = threading.RLock()
        self._stream_ttl = 1800
        self.acquisition_done = False
        self.counter = 0
        self.n_points = 0
        self.mca_names = [
            signal for signal in self.parent.component_names if signal.startswith("mca")
        ]
        self.mca_data = defaultdict(lambda: [])

    def initialize_detector(self) -> None:
        """Initialize detector"""
        # External trigger for pixel advance
        self.parent.channel_advance.set(ChannelAdvance.EXTERNAL)
        # Use internal clock for channel 1
        self.parent.channel1_source.set(ChannelAdvance.INTERNAL)
        self.parent.user_led.set(0)
        # Set number of channels to 5
        self.parent.mux_output.set(5)
        # Trigger Mode used for cSAXS
        self.parent.set_trigger(TriggerSource.MODE3)
        # specify polarity of trigger signals
        self.parent.input_polarity.set(0)
        self.parent.output_polarity.set(1)
        # do not start counting on start
        self.parent.count_on_start.set(0)
        self.stop_detector()

    def initialize_detector_backend(self) -> None:
        """Initialize detector backend"""
        for mca in self.mca_names:
            signal = getattr(self.parent, mca)
            signal.subscribe(self._on_mca_data, run=False)
        self.parent.current_channel.subscribe(self._progress_update, run=False)

    def _progress_update(self, value, **kwargs) -> None:
        """Progress update on the scan"""
        num_lines = self.parent.num_lines.get()
        max_value = self.parent.scaninfo.num_points
        # self.counter seems to be a deprecated variable from a former implementation of the mcs card
        # pylint: disable=protected-access
        self.parent._run_subs(
            sub_type=self.parent.SUB_PROGRESS,
            value=self.counter * int(self.parent.scaninfo.num_points / num_lines) + value,
            max_value=max_value,
            # TODO check if that is correct with
            done=bool(max_value == value),  # == self.counter),
        )

    @threadlocked
    def _on_mca_data(self, *args, obj=None, value=None, **kwargs) -> None:
        """Callback function for scan progress"""
        if not isinstance(value, (list, np.ndarray)):
            return
        self.mca_data[obj.attr_name] = value
        if len(self.mca_names) != len(self.mca_data):
            return
        self.acquisition_done = True
        self._send_data_to_bec()
        self.mca_data = defaultdict(lambda: [])

    def _send_data_to_bec(self) -> None:
        """Sends bundled data to BEC"""
        if self.parent.scaninfo.scan_msg is None:
            return
        metadata = self.parent.scaninfo.scan_msg.metadata
        metadata.update(
            {
                "async_update": "append",
                "num_lines": self.parent.num_lines.get(),
            }
        )
        msg = messages.DeviceMessage(
            signals=dict(self.mca_data),
            metadata=self.parent.scaninfo.scan_msg.metadata,
        )
        self.parent.connector.xadd(
            topic=MessageEndpoints.device_async_readback(
                scanID=self.parent.scaninfo.scanID, device=self.parent.name
            ),
            msg={"data": msg},
            expire=self._stream_ttl,
        )

    def prepare_detector(self) -> None:
        """Prepare detector for scan"""
        self.set_acquisition_params()
        self.parent.set_trigger(TriggerSource.MODE3)

    def set_acquisition_params(self) -> None:
        """Set acquisition parameters for scan"""
        if self.parent.scaninfo.scan_type == "step":
            self.n_points = int(self.parent.scaninfo.frames_per_trigger) * int(
                self.parent.scaninfo.num_points
            )
        elif self.parent.scaninfo.scan_type == "fly":
            self.n_points = int(self.parent.scaninfo.num_points)  # / int(self.num_lines.get()))
        else:
            raise MCSError(f"Scantype {self.parent.scaninfo} not implemented for MCS card")
        if self.n_points > 10000:
            raise MCSError(
                f"Requested number of points N={self.n_points} exceeds hardware limit of mcs card"
                " 10000 (N-1)"
            )
        self.parent.num_use_all.set(self.n_points)
        self.parent.preset_real.set(0)

    def prepare_detector_backend(self) -> None:
        """Prepare detector backend for scan"""
        self.parent.erase_all.set(1)
        self.parent.read_mode.set(ReadoutMode.EVENT)

    def arm_acquisition(self) -> None:
        """Arm detector for acquisition"""
        self.counter = 0
        self.parent.erase_start.set(1)

    def finished(self) -> None:
        """Check if acquisition is finished, if not successful, rais MCSTimeoutError"""
        signal_conditions = [
            (
                lambda: self.acquisition_done,
                True,
            ),
            (
                self.parent.acquiring.get,
                0,  # Considering making a enum.Int class for this state
            ),
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=True,
        ):
            total_frames = self.counter * int(
                self.parent.scaninfo.num_points / self.parent.num_lines.get()
            ) + max(self.parent.current_channel.get(), 0)
            raise MCSTimeoutError(
                f"Reached timeout with mcs in state {self.parent.acquiring.get()} and"
                f" {total_frames} frames arriving at the mcs card"
            )

    def stop_detector(self) -> None:
        """Stop detector"""
        self.parent.stop_all.set(1)

        return super().stop_detector()

    def stop_detector_backend(self) -> None:
        """Stop acquisition of data"""
        self.acquisition_done = True


class SIS38XX(Device):
    """SIS38XX card for access to EPICs PVs at cSAXS beamline"""


class MCScSAXS(PSIDetectorBase):
    """MCS card for cSAXS for implementation at cSAXS beamline"""

    USER_ACCESS = ["describe", "_init_mcs"]
    SUB_PROGRESS = "progress"
    SUB_VALUE = "value"
    _default_sub = SUB_VALUE

    # specify Setup class
    custom_prepare_cls = MCSSetup
    # specify minimum readout time for detector
    MIN_READOUT = 0

    # PV access to SISS38XX card
    # Acquisition
    erase_all = Cpt(EpicsSignal, "EraseAll")
    erase_start = Cpt(EpicsSignal, "EraseStart")  # ,trigger_value=1
    start_all = Cpt(EpicsSignal, "StartAll")
    stop_all = Cpt(EpicsSignal, "StopAll")
    acquiring = Cpt(EpicsSignal, "Acquiring")
    preset_real = Cpt(EpicsSignal, "PresetReal")
    elapsed_real = Cpt(EpicsSignal, "ElapsedReal")
    read_mode = Cpt(EpicsSignal, "ReadAll.SCAN")
    read_all = Cpt(EpicsSignal, "DoReadAll.VAL")  # ,trigger_value=1
    num_use_all = Cpt(EpicsSignal, "NuseAll")
    current_channel = Cpt(EpicsSignal, "CurrentChannel")
    dwell = Cpt(EpicsSignal, "Dwell")
    channel_advance = Cpt(EpicsSignal, "ChannelAdvance")
    count_on_start = Cpt(EpicsSignal, "CountOnStart")
    software_channel_advance = Cpt(EpicsSignal, "SoftwareChannelAdvance")
    channel1_source = Cpt(EpicsSignal, "Channel1Source")
    prescale = Cpt(EpicsSignal, "Prescale")
    enable_client_wait = Cpt(EpicsSignal, "EnableClientWait")
    client_wait = Cpt(EpicsSignal, "ClientWait")
    acquire_mode = Cpt(EpicsSignal, "AcquireMode")
    mux_output = Cpt(EpicsSignal, "MUXOutput")
    user_led = Cpt(EpicsSignal, "UserLED")
    input_mode = Cpt(EpicsSignal, "InputMode")
    input_polarity = Cpt(EpicsSignal, "InputPolarity")
    output_mode = Cpt(EpicsSignal, "OutputMode")
    output_polarity = Cpt(EpicsSignal, "OutputPolarity")
    model = Cpt(EpicsSignalRO, "Model", string=True)
    firmware = Cpt(EpicsSignalRO, "Firmware")
    max_channels = Cpt(EpicsSignalRO, "MaxChannels")

    # PV access to MCA signals
    mca1 = Cpt(EpicsSignalRO, "mca1.VAL", auto_monitor=True)
    mca3 = Cpt(EpicsSignalRO, "mca3.VAL", auto_monitor=True)
    mca4 = Cpt(EpicsSignalRO, "mca4.VAL", auto_monitor=True)
    current_channel = Cpt(EpicsSignalRO, "CurrentChannel", auto_monitor=True)

    # Custom signal readout from device config
    num_lines = Cpt(
        bec_utils.ConfigSignal,
        name="num_lines",
        kind="config",
        config_storage_name="mcs_config",
    )

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        device_manager=None,
        sim_mode=False,
        mcs_config=None,
        **kwargs,
    ):
        self.mcs_config = {
            f"{name}_num_lines": 1,
        }
        if mcs_config is not None:
            # pylint: disable=expression-not-assigned
            [self.mcs_config.update({f"{name}_{key}": value}) for key, value in mcs_config.items()]

        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            device_manager=device_manager,
            sim_mode=sim_mode,
            **kwargs,
        )

    def set_trigger(self, trigger_source: TriggerSource) -> None:
        """Set trigger mode from TriggerSource"""
        value = int(trigger_source)
        self.input_mode.set(value)

    def stage(self) -> list[object]:
        """stage the detector for upcoming acquisition"""
        rtr = super().stage()
        self.custom_prepare.arm_acquisition()
        return rtr


# Automatically connect to test environmenr if directly invoked
if __name__ == "__main__":
    mcs = MCScSAXS(name="mcs", prefix="X12SA-MCS:", sim_mode=True)



import time as ttime

from bec_lib import bec_logger
import numpy as np
from ophyd import Signal, Kind
from ophyd.utils import ReadOnlyError

logger = bec_logger.logger

# Readout precision for Setable/Readonly/ComputedReadonly signals
PRECISION = 3


class SetableSignal(Signal):
    """Setable signal for simulated devices.

    It will return the value of the readback signal based on the position
    created in the sim_state dictionary of the parent device.
    """

    def __init__(
        self,
        *args,
        name: str,
        value: any = None,
        kind: int = Kind.normal,
        precision: float = PRECISION,
        **kwargs,
    ):
        super().__init__(*args, name=name, value=value, kind=kind, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )
        self._value = value
        self.precision = precision
        # Init the sim_state, if self.parent.sim available, use it, else use self.parent
        self.sim = getattr(self.parent, "sim", self.parent)
        self._update_sim_state(value)

    def _update_sim_state(self, value: any) -> None:
        """Update the readback value."""
        self.sim.update_sim_state(self.name, value)

    def _get_value(self) -> any:
        """Update the timestamp of the readback value."""
        return self.sim.sim_state[self.name]["value"]

    def _get_timestamp(self) -> any:
        """Update the timestamp of the readback value."""
        return self.sim.sim_state[self.name]["timestamp"]

    def get(self):
        """Get the current position of the simulated device.

        Core function for signal.
        """
        self._value = self._get_value()
        return self._value

    def put(self, value):
        """Put the value to the simulated device.

        Core function for signal.
        """
        self._update_sim_state(value)
        self._value = value

    def describe(self):
        """Describe the readback signal.

        Core function for signal.
        """
        res = super().describe()
        if self.precision is not None:
            res[self.name]["precision"] = self.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self._get_timestamp()


class ReadOnlySignal(Signal):
    """Readonly signal for simulated devices.

    If initiated without a value, it will set the initial value to 0.
    """

    def __init__(
        self,
        *args,
        name: str,
        value: any = 0,
        kind: int = Kind.normal,
        precision: float = PRECISION,
        **kwargs,
    ):
        super().__init__(*args, name=name, value=value, kind=kind, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )
        self.precision = precision
        self._value = value
        # Init the sim_state, if self.parent.sim available, use it, else use self.parent
        self.sim = getattr(self.parent, "sim", None)
        self._init_sim_state()

    def _init_sim_state(self) -> None:
        """Init the readback value and timestamp in sim_state"""
        if self.sim:
            self.sim.update_sim_state(self.name, self._value)

    def _get_value(self) -> any:
        """Get the value of the readback from sim_state."""
        if self.sim:
            return self.sim.sim_state[self.name]["value"]
        else:
            return np.random.rand()

    def _get_timestamp(self) -> any:
        """Get the timestamp of the readback from sim_state."""
        if self.sim:
            return self.sim.sim_state[self.name]["timestamp"]
        else:
            return ttime.time()

    def get(self) -> any:
        """Get the current position of the simulated device.

        Core function for signal.
        """
        self._value = self._get_value()
        return self._value

    def put(self, value) -> None:
        """Put method, should raise ReadOnlyError since the signal is readonly."""
        raise ReadOnlyError(f"The signal {self.name} is readonly.")

    def describe(self):
        """Describe the readback signal.

        Core function for signal.
        """
        res = super().describe()
        if self.precision is not None:
            res[self.name]["precision"] = self.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self._get_timestamp()


class ComputedReadOnlySignal(Signal):
    """Computed readback signal for simulated devices.

    It will return the value computed from the sim_state of the signal.
    This can be configured in parent.sim.
    """

    def __init__(
        self,
        *args,
        name: str,
        value: any = None,
        kind: int = Kind.normal,
        precision: float = PRECISION,
        **kwargs,
    ):
        super().__init__(*args, name=name, value=value, kind=kind, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )
        self._value = value
        self.precision = precision
        # Init the sim_state, if self.parent.sim available, use it, else use self.parent
        self.sim = getattr(self.parent, "sim", self.parent)
        self._update_sim_state()

    def _update_sim_state(self) -> None:
        """Update the readback value.

        Call _compute_sim_state in parent device which updates the sim_state.
        """
        self.sim._compute_sim_state(self.name)

    def _get_value(self) -> any:
        """Update the timestamp of the readback value."""
        return self.sim.sim_state[self.name]["value"]

    def _get_timestamp(self) -> any:
        """Update the timestamp of the readback value."""
        return self.sim.sim_state[self.name]["timestamp"]

    def get(self):
        """Get the current position of the simulated device.

        Core function for signal.
        """
        self._update_sim_state()
        self._value = self._get_value()
        return self._value

    def put(self, value) -> None:
        """Put method, should raise ReadOnlyError since the signal is readonly."""
        raise ReadOnlyError(f"The signal {self.name} is readonly.")

    def describe(self):
        """Describe the readback signal.

        Core function for signal.
        """
        res = super().describe()
        if self.precision is not None:
            res[self.name]["precision"] = self.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self._get_timestamp()


if __name__ == "__main__":
    from ophyd_devices.sim import SimPositioner

    positioner = SimPositioner(name="positioner", parent=None)
    print(positioner.velocity.get())
    positioner.velocity.put(10)
    print(positioner.velocity.get())
    positioner.velocity.put(1)
    print(positioner.velocity.get())



import os
import time

from bec_lib import MessageEndpoints, bec_logger, messages
from ophyd import Component as Cpt
from ophyd import Device, DeviceStatus, EpicsSignal, EpicsSignalRO, Signal

logger = bec_logger.logger


class _SLSDetectorConfigSignal(Signal):
    def put(self, value, *, timestamp=None, force=False):
        self._readback = value
        self.parent.sim_state[self.name] = value

    def get(self):
        self._readback = self.parent.sim_state[self.name]
        return self.parent.sim_state[self.name]


# if (_eigerinvac_is_on == 1) {
#   tic("setup eiger in vac")
#   epics_put("X12SA-ES1-DOUBLE-02",0)
#     unix(sprintf("mkdir -p /sls/X12SA/data/%s/Data10/eiger_4/S%05d-%05d/S%05d",_username,int((SCAN_N+1)/1000)*1000,int((SCAN_N+1)/1000)*1000+999,SCAN_N+1))

#     epics_put("XOMNYI-DET-OUTDIR:0.DESC",sprintf("/sls/X12SA/data/%s/Data10/eiger_4/",_username))
#     epics_put("XOMNYI-DET-OUTDIR:1.DESC",sprintf("S%05d-%05d/S%05d",int((SCAN_N+1)/1000)*1000,int((SCAN_N+1)/1000)*1000+999,SCAN_N+1))
#    epics_put("XOMNYI-DET-CYCLES:0", _lamni_scan_numberofpts)
#     epics_put("XOMNYI-DET-EXPTIME:0", $2)
#     epics_put("XOMNYI-DET-INDEX:0", SCAN_N+1)

#     epics_put("XOMNYI-DET-CONTROL:0.DESC", "begin")
#     if(_eigerinvac_burst==0)
#     {
#       epics_put("XOMNYI-DET-CYCLES:0", _lamni_scan_numberofpts)
#       epics_put("XOMNYI-DET-EXPTIME:0", $2)
#       metadata_set("eiger_burst", "int", 1, 1)
#     }
#     else
#     {
#       metadata_set("eiger_burst", "int", 1, (int($2/0.0085)))
#       epics_put("XOMNYI-DET-CYCLES:0", _lamni_scan_numberofpts*(int($2/0.0085)))
#       epics_put("XOMNYI-DET-EXPTIME:0", 0.005)
#     }
#  global _DC_acquisition_ID
#     _DC_acquisition_ID = SCAN_N+1


class Eiger1p5MDetector(Device):
    USER_ACCESS = []
    file_path = Cpt(EpicsSignal, name="file_path", read_pv="XOMNYI-DET-OUTDIR:0.DESC")
    detector_out_scan_dir = Cpt(
        EpicsSignal, name="detector_out_scan_dir", read_pv="XOMNYI-DET-OUTDIR:1.DESC"
    )
    frames = Cpt(EpicsSignal, name="frames", read_pv="XOMNYI-DET-CYCLES:0")
    exp_time = Cpt(EpicsSignal, name="exp_time", read_pv="XOMNYI-DET-EXPTIME:0")
    index = Cpt(EpicsSignal, name="index", read_pv="XOMNYI-DET-INDEX:0")
    detector_control = Cpt(
        EpicsSignal, name="detector_control", read_pv="XOMNYI-DET-CONTROL:0.DESC"
    )
    framescaught = Cpt(EpicsSignalRO, name="framescaught", read_pv="XOMNYI-DET-CONTROL:0.VAL")

    file_pattern = Cpt(_SLSDetectorConfigSignal, name="file_pattern", value="")
    burst = Cpt(_SLSDetectorConfigSignal, name="burst", value=1)
    save_file = Cpt(_SLSDetectorConfigSignal, name="save_file", value=False)

    def __init__(self, *, name, kind=None, parent=None, device_manager=None, **kwargs):
        self.device_manager = device_manager
        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self.sim_state = {
            f"{self.name}_file_path": "~/Data10/data/",
            f"{self.name}_file_pattern": f"{self.name}_{{:05d}}.h5",
            f"{self.name}_frames": 1,
            f"{self.name}_burst": 1,
            f"{self.name}_save_file": False,
            f"{self.name}_exp_time": 0,
        }
        self._stopped = False
        self.file_name = ""
        self.metadata = {}
        self.username = "e20588"  # TODO get from config

    def _get_current_scan_msg(self) -> messages.ScanStatusMessage:
        return self.device_manager.connector.get(MessageEndpoints.scan_status())

    def _get_scan_dir(self, scan_bundle, scan_number, leading_zeros=None):
        if leading_zeros is None:
            leading_zeros = len(str(scan_bundle))
        floor_dir = scan_number // scan_bundle * scan_bundle
        return f"S{floor_dir:0{leading_zeros}d}-{floor_dir+scan_bundle-1:0{leading_zeros}d}/S{scan_number:0{leading_zeros}d}"

    def stage(self) -> list[object]:
        scan_msg = self._get_current_scan_msg()
        self.metadata = {
            "scanID": scan_msg.content["scanID"],
            "RID": scan_msg.content["info"]["RID"],
            "queueID": scan_msg.content["info"]["queueID"],
        }
        scan_number = scan_msg.content["info"]["scan_number"]
        exp_time = scan_msg.content["info"]["exp_time"]

        # set base path for detector output
        self.file_path.set(f"/sls/X12SA/data/{self.username}/Data10/eiger_4/")

        # set scan directory (e.g. S00000-00999/S00020)
        scan_dir = self._get_scan_dir(scan_bundle=1000, scan_number=scan_number, leading_zeros=5)
        self.detector_out_scan_dir.set(scan_dir)

        self.file_name = os.path.join(f"/sls/X12SA/data/{self.username}/Data10/eiger_4/", scan_dir)

        # set the scan number
        self.index.set(scan_number)

        # set the number of frames
        self.frames.set(scan_msg.content["info"]["num_points"])

        # set the exposure time
        self.exp_time.set(exp_time)

        # wait for detector control to become "ready"
        while True:
            det_ctrl = self.detector_control.get()
            if det_ctrl == "ready":
                break
            time.sleep(0.005)

        # send the "begin" flag to start processing the above commands
        self.detector_control.set("begin")

        # wait for detector to be "armed"
        logger.info("Waiting for detector setup")
        while True:
            det_ctrl = self.detector_control.get()
            if det_ctrl == "armed":
                break
            time.sleep(0.005)

        self.detector_control.put("acquiring")

        return super().stage()

    def unstage(self) -> list[object]:
        time_waited = 0
        sleep_time = 0.2
        framesexpected = self.frames.get()
        while True:
            framescaught = self.framescaught.get()
            if self.framescaught.get() < framesexpected:
                logger.info(
                    f"Waiting for frames: Transferred {framescaught} out of {framesexpected}"
                )
                time_waited += sleep_time
                time.sleep(sleep_time)
                if self._stopped:
                    break
                continue
            break
        self.detector_control.put("stop")
        signals = {"config": self.read(), "data": self.file_name}
        msg = messages.DeviceMessage(signals=signals, metadata=self.metadata)
        self.device_manager.connector.set_and_publish(MessageEndpoints.device_read(self.name), msg)
        self._stopped = False
        return super().unstage()

    def stop(self, *, success=False):
        self.detector_control.put("stop")
        super().stop(success=success)
        self._stopped = True


if __name__ == "__main__":
    eiger = Eiger1p5MDetector(name="eiger1p5m", label="eiger1p5m")
    breakpoint()



import time

from bec_lib import bec_logger
from bec_lib.devicemanager import DeviceContainer

from ophyd import Signal, Kind

from ophyd_devices.utils.socket import data_shape, data_type


logger = bec_logger.logger
DEFAULT_EPICSSIGNAL_VALUE = object()


# TODO maybe specify here that this DeviceMock is for usage in the DeviceServer
class DeviceMock:
    def __init__(self, name: str, value: float = 0.0):
        self.name = name
        self.read_buffer = value
        self._config = {"deviceConfig": {"limits": [-50, 50]}, "userParameter": None}
        self._enabled_set = True
        self._enabled = True

    def read(self):
        return {self.name: {"value": self.read_buffer}}

    def readback(self):
        return self.read_buffer

    @property
    def enabled_set(self) -> bool:
        return self._enabled_set

    @enabled_set.setter
    def enabled_set(self, val: bool):
        self._enabled_set = val

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, val: bool):
        self._enabled = val

    @property
    def user_parameter(self):
        return self._config["userParameter"]

    @property
    def obj(self):
        return self


class ProducerMock:
    def __init__(self, store_data=True) -> None:
        self.message_sent = []
        self._get_buffer = {}
        self.store_data = store_data

    def set(self, topic, msg, pipe=None, expire: int = None):
        if pipe:
            pipe._pipe_buffer.append(("set", (topic, msg), {"expire": expire}))
            return
        self.message_sent.append({"queue": topic, "msg": msg, "expire": expire})

    def send(self, topic, msg, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("send", (topic, msg), {}))
            return
        self.message_sent.append({"queue": topic, "msg": msg})

    def set_and_publish(self, topic, msg, pipe=None, expire: int = None):
        if pipe:
            pipe._pipe_buffer.append(("set_and_publish", (topic, msg), {"expire": expire}))
            return
        self.message_sent.append({"queue": topic, "msg": msg, "expire": expire})

    def lpush(self, topic, msg, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("lpush", (topic, msg), {}))
            return

    def rpush(self, topic, msg, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("rpush", (topic, msg), {}))
            return
        pass

    def lrange(self, topic, start, stop, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("lrange", (topic, start, stop), {}))
            return
        return []

    def get(self, topic, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("get", (topic,), {}))
            return
        val = self._get_buffer.get(topic)
        if isinstance(val, list):
            return val.pop(0)
        self._get_buffer.pop(topic, None)
        return val

    def keys(self, pattern: str) -> list:
        return []

    def pipeline(self):
        return PipelineMock(self)

    def delete(self, topic, pipe=None):
        if pipe:
            pipe._pipe_buffer.append(("delete", (topic,), {}))
            return

    def lset(self, topic: str, index: int, msgs: str, pipe=None) -> None:
        if pipe:
            pipe._pipe_buffer.append(("lrange", (topic, index, msgs), {}))
            return


class PipelineMock:
    _pipe_buffer = []
    _connector = None

    def __init__(self, connector) -> None:
        self._connector = connector

    def execute(self):
        if not self._connector.store_data:
            self._pipe_buffer = []
            return []
        res = [
            getattr(self._connector, method)(*args, **kwargs)
            for method, args, kwargs in self._pipe_buffer
        ]
        self._pipe_buffer = []
        return res


class DMMock:
    """Mock for DeviceManager

    The mocked DeviceManager creates a device containert and a connector.

    """

    def __init__(self):
        self.devices = DeviceContainer()
        self.connector = ProducerMock()

    def add_device(self, name: str, value: float = 0.0):
        self.devices[name] = DeviceMock(name, value)


class ConfigSignal(Signal):
    def __init__(
        self,
        *,
        name,
        value=0,
        timestamp=None,
        parent=None,
        labels=None,
        kind=Kind.hinted,
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
        config_storage_name: str = "config_storage",
    ):
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )

        self.storage_name = config_storage_name

    def get(self):
        self._readback = getattr(self.parent, self.storage_name)[self.name]
        return self._readback

    def put(
        self,
        value,
        connection_timeout=1,
        callback=None,
        timeout=1,
        **kwargs,
    ):
        """Using channel access, set the write PV to `value`.

        Keyword arguments are passed on to callbacks

        Parameters
        ----------
        value : any
            The value to set
        connection_timeout : float, optional
            If not already connected, allow up to `connection_timeout` seconds
            for the connection to complete.
        use_complete : bool, optional
            Override put completion settings
        callback : callable
            Callback for when the put has completed
        timeout : float, optional
            Timeout before assuming that put has failed. (Only relevant if
            put completion is used.)
        """

        old_value = self.get()
        timestamp = time.time()
        getattr(self.parent, self.storage_name)[self.name] = value
        super().put(value, timestamp=timestamp, force=True)
        self._run_subs(
            sub_type=self.SUB_VALUE,
            old_value=old_value,
            value=value,
            timestamp=timestamp,
        )

    def describe(self):
        """Provide schema and meta-data for :meth:`~BlueskyInterface.read`

        This keys in the `OrderedDict` this method returns must match the
        keys in the `OrderedDict` return by :meth:`~BlueskyInterface.read`.

        This provides schema related information, (ex shape, dtype), the
        source (ex PV name), and if available, units, limits, precision etc.

        Returns
        -------
        data_keys : OrderedDict
            The keys must be strings and the values must be dict-like
            with the ``event_model.event_descriptor.data_key`` schema.
        """
        if self._readback is DEFAULT_EPICSSIGNAL_VALUE:
            val = self.get()
        else:
            val = self._readback
        return {
            self.name: {
                "source": f"{self.parent.prefix}:{self.name}",
                "dtype": data_type(val),
                "shape": data_shape(val),
            }
        }
